/*
 * wtmpclean.c -- clear from a wtmp database the unwanted records
 * Copyright (C) 2008 by Davide Madrisan <davide.madrisan@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * indent -i4 -nut src/wtmpclean.c
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif
#ifdef HAVE_STRING_H
# if !defined STDC_HEADERS && defined HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif

#include <errno.h>
#include <unistd.h>
#include <utime.h>

#ifdef HAVE_UTMP_H
# include <utmp.h>
#endif

#include "wtmpclean.h"

#if !HAVE_MEMSET
extern void *memset (void *s, int c, int n);
#endif

void
usage (int status)
{
    static const char *banner[] = {
        PACKAGE " version " PACKAGE_VERSION,
        "Copyright (C) 2008 by Davide Madrisan <davide.madrisan@gmail.com>",
        "",
        "Usage: " PACKAGE " -u <user> [-p] [-f wtmpfile]",
        "  -f   modify <wtmpfile> instead of " _PATH_WTMP,
        "  -p   preserve ownership and timestamps",
        "  -t   just check for <user> logins",
        "  -u   remove all the blocks concerning <user>",
        "",
        "This is free software; see the source for copying conditions.  There is NO",
        "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.",
        (char *) 0
    };
    unsigned int i;

    for (i = 0; banner[i]; i++)
      {
          fprintf (stdout, "%s\n", banner[i]);
      }
    exit (status);
}

int
main (int argc, char **argv)
{
    int opt;
    unsigned int tflag = 0, pflag = 0;
    char *wtmpfile = _PATH_WTMP;
    char *user = NULL;

    opterr = 0;
    while ((opt = getopt (argc, argv, "f:tu:hp")) != -1)
      {
          switch (opt)
            {
            case 'f':
                wtmpfile = optarg;
                break;
            case 'p':
                pflag = 1;
                break;
            case 't':
                tflag = 1;
                break;
            case 'u':
                user = optarg;
                break;
            default:
                usage ((opt == 'h') ? 1 : 0);
            }
      }

    if (!user)
      {
          usage (1);
      }

    if (tflag)
      {
          printf ("%u entries for user %s found\n",
                  wtmpchkusr (wtmpfile, user), user);
          exit (0);
      }

    wtmpedit (wtmpfile, user, pflag);
}

void
wtmpedit (const char *wtmpfile, const char *user, unsigned int preserve)
{
    struct utmp *utp;
    unsigned int usrrec, cleanok, cleanko;
    struct stat sb;
    struct utimbuf time;
    uid_t owner;
    gid_t group;

    if (preserve)
      {
          if (stat (wtmpfile, &sb) < 0)
            {
                perror ("Cannot get file status");
                exit (1);
            }
          time.actime = sb.st_atime;
          time.modtime = sb.st_mtime;
          owner = sb.st_uid;
          group = sb.st_gid;
      }

    utmpname (wtmpfile);
    setutent ();

    usrrec = cleanok = cleanko = 0;
    while (utp = getutent ())
      {
          if (utp->ut_type == USER_PROCESS &&
              strncmp (utp->ut_user, user, sizeof utp->ut_user) == 0)
            {
                utp->ut_type = DEAD_PROCESS;
                memset (utp->ut_line, 0, sizeof utp->ut_line);
                utp->ut_time = 0;
                memset (utp->ut_user, 0, sizeof utp->ut_user);

                (utp = pututline (utp)) ? cleanok++ : cleanko++;
                usrrec++;
            }
      }

    if (cleanok > 0)
      {
          fprintf (stderr, "Cannot clean up %s\n", wtmpfile);
      }
    else
      {
          printf ("%u block(s) logging user %s removed from %s\n",
                  usrrec, user, wtmpfile);
      }

    endutent ();

    if (preserve && chown (wtmpfile, owner, group) < 0)
      {
          fprintf (stderr,
                   "Cannot restore ownership of %s: %s\n",
                   wtmpfile, strerror (errno));
      }
    if (preserve && utime (wtmpfile, &time) < 0)
      {
          fprintf (stderr,
                   "Cannot restore access and modification times of %s: %s\n",
                   wtmpfile, strerror (errno));
      }
}

unsigned int
wtmpchkusr (const char *wtmpfile, const char *user)
{
    struct utmp *utp;
    unsigned int found = 0;

    utmpname (wtmpfile);
    setutent ();

    while (utp = getutent ())
      {
          if (utp->ut_type == USER_PROCESS &&
              strncmp (utp->ut_user, user, sizeof utp->ut_user) == 0)
            {
                found++;
            }
      }

    endutent ();

    return found;
}
